from rest_framework import mixins
from rest_framework_mongoengine.viewsets import GenericViewSet as MEGenericViewSet
from rest_framework_mongoengine import viewsets as mongoviewsets

from . import utils
from .. import models
from .. import serializers


class USLViewSet(mixins.RetrieveModelMixin,
                 mixins.ListModelMixin,
                 MEGenericViewSet):
    queryset = models.USL.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.USLSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.USL, html=html)

    def get_queryset(self):
        return models.USL.objects.all()


class WordViewSet(mongoviewsets.ModelViewSet):
    queryset = models.Word.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.WordSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.Word, html=html)

    def get_queryset(self):
        return models.Word.objects.all()


class TextViewSet(mongoviewsets.ModelViewSet):
    queryset = models.Text.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.TextSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.Text, html=html)

    def get_queryset(self):
        return models.Text.objects.all()



from mongoengine import Document, fields


class USL(Document):
    ieml_text = fields.StringField(unique=True, required=True,)
    # Can be blank because a USL attached to a post is likely to be the
    # union of the tags, with no particular translation in natural language.
    translations = fields.MapField(
        fields.StringField(),
        help_text=('Translations in natural languages. '
                   'Use the ISO 639-1 format for the keys.'),
    )

    meta = {
        'ordering': ['ieml_text'],
        'indexes': [
            'ieml_text',
        ],
        'allow_inheritance': True,
    }

    def __str__(self):
        return '{0}: {1}'.format(
            self.__class__.__name__,
            self.ieml_text,
        )



from django.utils import timezone
from mongoengine import Document, fields, signals


class Collection(Document):
    """A coherent group of documents gathered by a user."""

    title = fields.StringField(unique=True, required=True,)
    curators = fields.ListField(fields.StringField(), required=True,)
    created_on = fields.DateTimeField(default=timezone.now,)
    updated_on = fields.DateTimeField(default=timezone.now,)
    # Use a MapField instead of a ListField to ensure that a document is not
    # collected more than once
    posts = fields.MapField(
        fields.EmbeddedDocumentField('Post'),
        help_text="The key must be equal to the 'document' field of the post.",
    )
    # Use a MapField to retrieve a source from its id efficiently.
    # Also make sure that ids are unique.
    sources = fields.MapField(
        fields.EmbeddedDocumentField('CollectedSource'),
        help_text="The key must be equal to the 'id' field of the collected source.",
    )

    meta = {
        'ordering': ['title'],
    }

    def __str__(self):
        return self.title

    @classmethod
    def pre_save(cls, sender, document, **kwargs):
        document.updated_on = timezone.now()

        for key, source in document.sources.items():
            if key:
                continue

            document.sources.pop(key)
            key = source.generate_key()
            source.id = key
            document.sources[key] = source


signals.pre_save.connect(Collection.pre_save, sender=Collection)



from django.utils.encoding import smart_text
from rest_framework.utils import formatting


def get_modelview_description(model, html=False):
    description = model.__doc__ or ''
    description = formatting.dedent(smart_text(description))
    if html:
        return formatting.markup_description(description)
    return description



import uuid

from mongoengine import EmbeddedDocument, fields


class CollectedSource(EmbeddedDocument):
    """A request pattern for a source driver. For instance, all tweets having
    a particular tag. The posts to collect are described in the `params` field,
    the content of which depends on the driver API.
    """

    # To have an immutable identifier for the source
    id = fields.StringField(required=True,)
    driver = fields.ReferenceField('SourceDriver', required=True,)
    params = fields.DictField(
        # Cannot be blank because scrapers need to know which posts to collect
        # on the source
        required=True,
        help_text='A dictionnary of parameters passed to the driver.',
    )
    last_request_date = fields.DateTimeField(null=True,)
    being_requested = fields.BooleanField(default=False,)

    @staticmethod
    def generate_key():
        return uuid.uuid4().hex



# TODO: to be removed, for testing ui
from django.shortcuts import render
from django.views.decorators.http import require_http_methods
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse

def home(request):
    return render(request, 'intlekt/home.html')

def usl_tagger(request):
    return render(request, 'intlekt/uslTagger.html')

@require_http_methods(["POST"])
@csrf_exempt
def scoopit(request):
    print(request.body)

    return HttpResponse('scoopit')


from . import collections
from . import library



from mongoengine import Document, fields


class SourceDriver(Document):
    """A service to pull posts from a source."""

    source = fields.ReferenceField('Source', required=True,)
    url = fields.URLField(unique=True, required=True,)



from mongoengine import Document as MEDocument, fields


class Document(MEDocument):
    """A document on the web.

    Most fields can be null because scrapers may not be able to detect them
    automatically. Users will edit them afterwards.
    """

    title = fields.StringField(
        null=True,
        help_text='Null if unknown.',
    )
    authors = fields.ListField(fields.StringField(),)
    created_on = fields.DateTimeField(
        null=True,
        help_text='Null if unknown.',
    )
    # Not unique because a single web page can contain multiple documents
    url = fields.URLField(required=True,)
    keywords = fields.ListField(fields.StringField(),)
    language = fields.StringField(
        null=True,
        help_text='Use the ISO 639-1 format. Null if unknown.',
    )

    def __str__(self):
        return self.title



from django.utils import timezone
from mongoengine import EmbeddedDocument, fields

from .utils import USLField


class Post(EmbeddedDocument):
    """A collected document, i.e. a document with metadata to be used for
    indexing. For instance, a post can be a tweet refering to an article.
    """

    document = fields.ReferenceField('Document', required=True,)
    collected_on = fields.DateTimeField(required=True,)
    # Can be null because finding a correct usl is complex so the user may not
    # do it immediatly, and so do scrapers.
    usl = USLField(null=True, default=None,)
    tags = fields.ListField(fields.StringField(),)
    # For display purposes
    image = fields.URLField(verify_exists=True, null=True, default=None,)
    url = fields.URLField(null=True, default=None,)
    description = fields.StringField(default="",)
    # To mark a document as irrelevant for the collection. Do not delete it to
    # keep track of its past presence.
    hidden = fields.BooleanField(default=False,)



import pycountry

from rest_framework import serializers


def validate_set(list_):
    if len(set(list_)) != len(list_):
        raise serializers.ValidationError('Cannot have duplicates.')


def validate_language(language):
    try:
        pycountry.languages.get(alpha_2=language)
    except KeyError:
        raise serializers.ValidationError('Bad format. See ISO 639-1.')



from mongoengine import Document, fields

from .utils import USLField


class Tag(Document):
    """An IEML-translated word used to tag documents."""

    # The meaning can depend on the context, so a tag may have multiple usls.
    # Cannot be empty for the Mongo collection to contain translated tags only.
    # If it could be, there would be two ways to mark a tag as not translated:
    # not in the Mongo collection or with its `usls` attribute empty. It would
    # introduce additional parsing.
    usls = fields.ListField(USLField(required=True,), required=True,)
    text = fields.StringField(unique=True, required=True,)

    meta = {
        'ordering': ['text'],
    }



from rest_framework_mongoengine import serializers as mongoserializers
from rest_framework import serializers

from . import utils
from .. import models


def validate_translations_dict(translations):
    for lang in translations:
        utils.validate_language(lang)


def validate_text_dependencies_number(dependencies):
    # If there is one dependency only, the objects are considered equal.
    # For instance, a text with one word is the word itself.
    if len(dependencies) < 2:
        raise serializers.ValidationError('Must have two dependencies or more.')


def validate_text_dependencies_non_recursivity(dependencies):
    # A text cannot depend on another text. To include a text into another,
    # just copy all its dependencies.
    ids = [dep.id for dep in dependencies]
    usls = models.Text.objects(id__in=ids)
    if usls:
        raise serializers.ValidationError(
            'A text cannot contain another text. '
            'Instead, copy all its dependencies.'
        )


def serializer_factory(name, Model, extra_kwargs=None):
    extra_kwargs_ = extra_kwargs or {}

    class Meta:
        model = Model
        exclude = ('_cls',)
        extra_kwargs = {
            'translations': {'validators': [validate_translations_dict]},
            **extra_kwargs_,
        }

    return type(name, (mongoserializers.DocumentSerializer,), {'Meta': Meta})


class USLSerializer(mongoserializers.DocumentSerializer):
    class Meta:
        model = models.USL
        fields = '__all__'


TextSerializer = serializer_factory(
    'TextSerializer',
    models.Text,
    extra_kwargs={
        'dependencies': {
            'validators': [
                utils.validate_set,
                validate_text_dependencies_number,
                validate_text_dependencies_non_recursivity,
            ]
        }
    }
)

WordSerializer = serializer_factory('WordSerializer', models.Word)



from django.conf.urls import include, url
from rest_framework_mongoengine.routers import DefaultRouter
from rest_framework_nested.routers import NestedSimpleRouter

from . import views


router = DefaultRouter()
router.register(r'collections', views.collections.CollectionViewSet)
router.register(r'documents', views.collections.DocumentViewSet)
router.register(r'sources', views.collections.SourceViewSet)
router.register(r'source_drivers', views.collections.SourceDriverViewSet)
router.register(r'tags', views.collections.TagViewSet)
router.register(r'texts', views.library.TextViewSet)
router.register(r'usls', views.library.USLViewSet)
router.register(r'words', views.library.WordViewSet)

posts_router = NestedSimpleRouter(router, r'collections', lookup='collection')
posts_router.register(
    r'posts',
    views.collections.PostViewSet,
    base_name='collection_posts',
)

collected_sources_router = NestedSimpleRouter(router, r'collections', lookup='collection')
collected_sources_router.register(
    r'sources',
    views.collections.CollectedSourceViewSet,
    base_name='collection_collected_sources',
)


urlpatterns = [
    url(r'^', include(router.urls)),
    url(r'^', include(posts_router.urls)),
    url(r'^', include(collected_sources_router.urls)),

    # TODO: to be removed, for testing
    url(r'^ui/$', views.home),
    url(r'^ui/usl_tagger$', views.usl_tagger),
    url(r'^scoopit/$', views.scoopit),
]



from .collections.collection import Collection
from .collections.collected_source import CollectedSource
from .collections.document import Document
from .collections.post import Post
from .collections.source import Source
from .collections.source_driver import SourceDriver
from .collections.tag import Tag

from .library.text import Text
from .library.usl import USL
from .library.word import Word



from django.utils.encoding import smart_text
from rest_framework_mongoengine import viewsets as mongoviewsets
from rest_framework import viewsets
from rest_framework import status
from rest_framework.response import Response
from rest_framework.utils import formatting
from rest_framework.decorators import detail_route

from . import utils
from .. import models
from .. import serializers


class NestedObjectDoesNotExist(LookupError): pass


class NestedCollectionResourceViewSet(viewsets.ViewSet):
    serializer_class = None
    model = None
    resource_name = None

    @staticmethod
    def nested_obj_to_key(obj):
        raise NotImplementedError()

    def get_serializer(self, *args, **kwargs):
        return self.serializer_class(*args, **kwargs)

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(cls.model, html=html)

    @classmethod
    def get_nested_object(cls, collection_id, object_id):
        try:
            collection = models.Collection.objects.get(id=collection_id)
        except models.Collection.DoesNotExist:
            raise NestedObjectDoesNotExist()

        objects = getattr(collection, cls.resource_name)
        try:
            return objects[object_id], collection
        except KeyError:
            raise NestedObjectDoesNotExist()

    def list(self, request, collection_id=None):
        if collection_id is None:
            return Response('Please specify a collection id.',
                            status=status.HTTP_400_BAD_REQUEST)

        try:
            collection = models.Collection.objects.get(id=collection_id)
        except models.Collection.DoesNotExist:
            return Response('No such collection {}'.format(collection_id),
                            status=status.HTTP_404_NOT_FOUND)
        objects = getattr(collection, self.resource_name)

        for key in objects:
            objects[key] = self.serializer_class(objects[key]).data

        return Response(objects)

    def create(self, request, collection_id=None):
        if collection_id is None:
            return Response('Please specify a collection id.',
                            status=status.HTTP_400_BAD_REQUEST)

        try:
            collection = models.Collection.objects.get(id=collection_id)
        except models.Collection.DoesNotExist:
            return Response('No such collection {}'.format(collection_id),
                            status=status.HTTP_404_NOT_FOUND)

        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors,
                            status=status.HTTP_400_BAD_REQUEST)

        obj = serializer.save()

        if self.nested_obj_to_key(obj) in getattr(collection, self.resource_name):
            return Response(
                {self.resource_name: ('This document has already been collected. '
                              'Please, use PUT or PATCH.')},
                status=status.HTTP_400_BAD_REQUEST
            )

        getattr(collection, self.resource_name)[self.nested_obj_to_key(obj)] = obj
        collection.save()

        return Response(serializer.data)

    def retrieve(self, request, pk=None, collection_id=None):
        try:
            obj, _ = self.get_nested_object(collection_id, pk)
        except NestedObjectDoesNotExist:
            return Response(
                'No such object {} in collection {} {}'.format(
                    pk,
                    collection_id,
                    self.resource_name,
                ),
                status=status.HTTP_404_NOT_FOUND
            )

        return Response(self.serializer_class(obj).data)

    def update(self, request, pk=None, collection_id=None):
        try:
            obj, collection = self.get_nested_object(collection_id, pk)
        except NestedObjectDoesNotExist:
            return Response(
                'No such object {} in collection {} {}'.format(
                    pk,
                    collection_id,
                    self.resource_name,
                ),
                status=status.HTTP_404_NOT_FOUND
            )

        serializer = self.serializer_class(obj, data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors,
                            status=status.HTTP_400_BAD_REQUEST)

        obj = serializer.save()
        getattr(collection, self.resource_name)[pk] = obj
        collection.save()

        return Response(serializer.data)

    def partial_update(self, request, pk=None, collection_id=None):
        try:
            obj, collection = self.get_nested_object(collection_id, pk)
        except NestedObjectDoesNotExist:
            return Response(
                'No such object {} in collection {} {}'.format(
                    pk,
                    collection_id,
                    self.resource_name,
                ),
                status=status.HTTP_404_NOT_FOUND
            )

        serializer = self.serializer_class(
            obj,
            data=request.data,
            partial=True,
        )
        if not serializer.is_valid():
            return Response(serializer.errors,
                            status=status.HTTP_400_BAD_REQUEST)

        obj = serializer.save()
        getattr(collection, self.resource_name)[pk] = obj
        collection.save()

        return Response(serializer.data)

    def destroy(self, request, pk=None, collection_id=None):
        try:
            _, collection = self.get_nested_object(collection_id, pk)
        except NestedObjectDoesNotExist:
            return Response(
                'No such object {} in collection {} {}'.format(
                    pk,
                    collection_id,
                    self.resource_name
                ),
                status=status.HTTP_404_NOT_FOUND
            )

        getattr(collection, self.resource_name).pop(pk, None)
        collection.save()

        return Response('')


class PostViewSet(NestedCollectionResourceViewSet):
    serializer_class = serializers.PostSerializer
    model = models.Post
    resource_name = 'posts'

    @staticmethod
    def nested_obj_to_key(post):
        return str(post.document.id)


class CollectedSourceViewSet(NestedCollectionResourceViewSet):
    serializer_class = serializers.CollectedSourceSerializer
    model = models.CollectedSource
    resource_name = 'sources'

    @staticmethod
    def nested_obj_to_key(source):
        # If the source is created source.id == None, which is not a valid
        # dict key
        return source.id or ''


class CollectionViewSet(mongoviewsets.ModelViewSet):
    queryset = models.Collection.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.CollectionSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.Collection, html=html)

    def get_queryset(self):
        return models.Collection.objects.all()


class DocumentViewSet(mongoviewsets.ModelViewSet):
    queryset = models.Document.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.DocumentSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.Document, html=html)

    def get_queryset(self):
        return models.Document.objects.all()


class TagViewSet(mongoviewsets.ModelViewSet):
    queryset = models.Tag.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.TagSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.Tag, html=html)

    def get_queryset(self):
        return models.Tag.objects.all()


class SourceViewSet(mongoviewsets.ModelViewSet):
    queryset = models.Source.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.SourceSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.Source, html=html)

    def get_queryset(self):
        return models.Source.objects.all()


class SourceDriverViewSet(mongoviewsets.ModelViewSet):
    queryset = models.SourceDriver.objects.all()
    lookup_field = 'id'
    serializer_class = serializers.SourceDriverSerializer

    @classmethod
    def get_view_description(cls, html=False):
        return utils.get_modelview_description(models.SourceDriver, html=html)

    def get_queryset(self):
        return models.SourceDriver.objects.all()



from rest_framework_mongoengine import serializers as mongoserializers
from rest_framework import serializers

from . import utils
from .. import models


def validate_sources_dict(sources):
    for key, source in sources.items():
        if key != source['id']:
            raise serializers.ValidationError(
                "The key '{0}' must be equal to the 'id' field value "
                "'{1}'.".format(key, source['id'])
            )


def validate_posts_dict(posts):
    for key, post in posts.items():
        if key != str(post['document'].id):
            raise serializers.ValidationError(
                "The key '{0}' must be equal to the 'document' field value "
                "'{1}'.".format(key, post['document'].id)
            )


class CollectedSourceSerializer(mongoserializers.EmbeddedDocumentSerializer):
    class Meta:
        model = models.CollectedSource
        fields = '__all__'
        read_only_fields = ('id',) 
        extra_kwargs = {
            'last_request_date': {
                'format': '%Y-%m-%d',
                'input_formats': ['%Y-%m-%d'],
            },
        }


class PostSerializer(mongoserializers.EmbeddedDocumentSerializer):
    class Meta:
        model = models.Post
        fields = '__all__'
        extra_kwargs = {
            'collected_on': {
                'format': '%Y-%m-%d',
                'input_formats': ['%Y-%m-%d'],
            },
            'description': {'allow_blank': True},
            'tags': {'validators': [utils.validate_set]},
        }


class CollectionSerializer(mongoserializers.DocumentSerializer):
    class Meta:
        model = models.Collection
        fields = '__all__'
        read_only_fields = ('created_on', 'updated_on',)
        extra_kwargs = {
            'created_on': {'format': '%Y-%m-%d'},
            'updated_on': {'format': '%Y-%m-%d'},
            'authors': {'validators': [utils.validate_set]},
            'posts': {
                'child': PostSerializer(),
                'validators': [validate_posts_dict],
            },
            'sources': {
                'child': CollectedSourceSerializer(),
                'validators': [validate_sources_dict],
            },
        }


class DocumentSerializer(mongoserializers.DocumentSerializer):
    class Meta:
        model = models.Document
        fields = '__all__'
        extra_kwargs = {
            'created_on': {
                'format': '%Y-%m-%d',
                'input_formats': ['%Y-%m-%d'],
            },
            'authors': {'validators': [utils.validate_set]},
            'keywords': {'validators': [utils.validate_set]},
            'language': {'validators': [utils.validate_language]},
            'title': {'allow_blank': True},
        }


class TagSerializer(mongoserializers.DocumentSerializer):
    class Meta:
        model = models.Tag
        fields = '__all__'
        extra_kwargs = {
            'usls': {'validators': [utils.validate_set]},
        }


class SourceSerializer(mongoserializers.DocumentSerializer):
    class Meta:
        model = models.Source
        fields = '__all__'


class SourceDriverSerializer(mongoserializers.DocumentSerializer):
    class Meta:
        model = models.SourceDriver
        fields = '__all__'



from mongoengine import Document, fields


class Source(Document):
    """A source from which posts can be pulled. For instance, Twitter.
    
    Use a model instead of a simple `StringField` to avoid duplicates like
    "Twitter", "twitter", "touiteur".
    """

    name = fields.StringField(unique=True, required=True,)



import os
from setuptools import setup, find_packages


def readme():
    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
        return f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-intlekt',
    version='0.1',
    url='https://github.com/IEMLdev/django-intlekt',
    description='Django app for Intlekt, a content curation application using the IEML language.',
    long_description=readme(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Text Processing :: Indexing'
    ],
    author='Vincent Lefoulon',
    author_email='vincent.lefoulon@free.fr',
    keywords='ieml intlekt semantic syntax language indexing',
    license='GPLv3',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'mongoengine',
        'blinker',  # For Mongoengine signals
        'django',
        'djangorestframework',
        'django-rest-framework-mongoengine',
        'drf-nested-routers',
        'pycountry',
    ],
)



from .collections import (
    PostSerializer, CollectedSourceSerializer, CollectionSerializer, DocumentSerializer,
    TagSerializer, SourceSerializer, SourceDriverSerializer
)
from .library import USLSerializer, WordSerializer, TextSerializer
